package CH.psi.midas;

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

public class Experiment extends java.lang.Object implements ActionListener,Serializable
{
  public static final int MIDAS_PORT = 1175;
  public static final String MIDAS_VERSION = "1.7.0";

  public static final int MERROR = (1<<0);
  public static final int MINFO  = (1<<1);
  public static final int MDEBUG = (1<<2);
  public static final int MUSER  = (1<<3);
  public static final int MLOG   = (1<<4);
  public static final int MTALK  = (1<<5);
  public static final int MCALL  = (1<<6);
  public static final int MALL   = 0xFF;

  private String host = "";
  private String experiment = "";
  private String client = "";
  
  private Socket sockSend, sockRecv, sockEvent;
  private PrintWriter pw;
  private InputStream is, recvIs;

  private static final int RPC_TIMEOUT = 10000;
  
  private int hDB = 0;
  private int hBufMsg, hBufEvent;
  
  private Timer  timerHotLink;
  private Vector vecHotLink = new Vector();
  
  private MessageListener messageListener = null;
  
  public Experiment()
  {
  }
  
  public Experiment(String host, String experiment, String client)
  {
    this.host = host;
    this.experiment = experiment;
    this.client = client;
  }

  public void setHost(java.lang.String host)
  {
    this.host = host;
  }

  public java.lang.String getHost()
  {
    return this.host;
  }

  public void setExperiment(String experiment)
  {
    this.experiment = experiment;
  }

  public java.lang.String getExperiment()
  {
    return this.experiment;
  }

  public void setClient(String client)
  {
    this.client = client;
  }

  public java.lang.String getClient()
  {
    return this.client;
  }

  public int getHdb() 
  {
    return this.hDB;
  }

  public void connect() throws UnknownExpException, IOException
  {
    // use local host as default
    if (host.equals(""))
      host = java.net.InetAddress.getLocalHost().getHostName();
    
    // create server sockets
    ServerSocket s1 = new ServerSocket(0);
    ServerSocket s2 = new ServerSocket(0);
    ServerSocket s3 = new ServerSocket(0);
      
    Socket s = new Socket(host, MIDAS_PORT);
      
    is = s.getInputStream();
    pw = new PrintWriter(s.getOutputStream());
      
    // send connect request
    if (experiment.length() > 0)
      pw.print("C "+Integer.toString(s1.getLocalPort())+" "+
                      Integer.toString(s2.getLocalPort())+" "+
                      Integer.toString(s3.getLocalPort())+" "+
                      MIDAS_VERSION+" "+experiment+(char) 0);
    else                        
      pw.print("C "+Integer.toString(s1.getLocalPort())+" "+
                      Integer.toString(s2.getLocalPort())+" "+
                      Integer.toString(s3.getLocalPort())+" "+
                      MIDAS_VERSION+" Default"+(char) 0);
    pw.flush();

    // receive reply
    String reply = "";
    int i;
    s.setSoTimeout(RPC_TIMEOUT);
    do
      {
      i = is.read();  
      if (i == -1)
        break;
          
      reply += (char) i;
      } while (i != 0);
    s.close();
    
    if (reply.charAt(0) == '2')
      throw new UnknownExpException("Experiment "+experiment);
      
    // check version
    StringTokenizer st = new StringTokenizer(reply.substring(2), " .");
    StringTokenizer vst = new StringTokenizer(MIDAS_VERSION, ".");
    if (!st.nextToken().equals(vst.nextToken()) ||
        !st.nextToken().equals(vst.nextToken()))
      JOptionPane.showMessageDialog(null, "Server MIDAS version ("+reply.substring(2).trim()+ 
                                          ") differs from local version ("+MIDAS_VERSION+")", 
                                          "JMIDAS" , 
                                          JOptionPane.ERROR_MESSAGE, 
                                          null);
              
    // accept callback
    sockSend  = s1.accept();
    sockRecv  = s2.accept();
    sockEvent = s3.accept();
    
    s1.close();
    s2.close();
    s3.close();

    // set timeouts
    sockSend.setSoTimeout(RPC_TIMEOUT); // for RPC reply
    sockRecv.setSoTimeout(1);           // for hotlink notification

    pw = new PrintWriter(sockSend.getOutputStream());
    is = sockSend.getInputStream();
    recvIs = sockRecv.getInputStream();

    // send computer info
    pw.print("128 "+client+(char)0);
    pw.flush();

    // receive reply
    reply = "";
    do
      {
      i = is.read();  
      if (i == -1)
        break;
          
      reply += (char) i;
      } while (i != 0);
      
    // open experiment database
    reply = rpcCall("db_open_database&ODB&100000&"+client);
    hDB = Integer.parseInt(replyParam(reply, 1));

    // set client info in ODB
    reply = rpcCall("cm_set_client_info&"+hDB+"&"+InetAddress.getLocalHost().getHostName()+"&"+client+"&128&"+"");
    while (replyParam(reply, 0).equals("108")) {
      JPasswordField pw = new JPasswordField();
        
      int ok = JOptionPane.showConfirmDialog(null, 
                                             pw, 
                                             "Enter password of experiment" , 
                                             JOptionPane.OK_CANCEL_OPTION, 
                                             JOptionPane.PLAIN_MESSAGE);
      if (ok != JOptionPane.OK_OPTION)
        throw new IOException("Connection cancelled");
 
      // reopen database
      hDB = Integer.parseInt(replyParam(rpcCall("db_open_database&ODB&100000&"+client), 1));

      String pwd = crypt(pw.getPassword().toString(), "mi");
      
      reply = rpcCall("cm_set_client_info&"+hDB+"&"+
                       InetAddress.getLocalHost().getHostName()+"&"+client+"&128&"+pwd);
    }
    
  // Send start notification  
  messageSend(MINFO, "Program "+client+" on host "+
              InetAddress.getLocalHost().getHostName()+" started");
        
  // Create and start hotlink timer
  timerHotLink = new Timer(100, this);
  timerHotLink.start();
  }
  
  public synchronized String rpcCall(String arg) throws IOException {
    int n;
    byte b[] = new byte[10000];
    String reply = "";
    
    // send arguments
    pw.print(arg+(char)0);
    pw.flush();
    
    // wait for reply
    do {
      if (is.available() > 0) {
        n = is.read(b);  

        // break if connection broken
        if (n == -1 || n == 0)
          break;
        
        if (b[n-1] == 0) {
          // break if end of string received  
          reply += new String(b, 0, n-1);
          break;
        } else
          // add partial reply
          reply += new String(b, 0, n);

      } else 
          try { Thread.sleep(10); } catch (Exception e) {}
  
    } while (true);

    return reply;
  }

  public synchronized byte[] rpcReceiveEvent(int hBuf) throws IOException {

    int i;
    
    // send rpc command
    pw.print("bm_receive_event&"+hBuf+"&10000&1"+(char)0);
    pw.flush();
    
    // receive reply
    String reply = "";
    do
      {
      i = is.read();  
      if (i == -1 || i == 0)
        break;
          
      reply += (char) i;
      } while (true);
    
    int n = Integer.parseInt(replyParam(reply, 1));
    
    // receive binary event
    byte array[] = new byte[n];

    if (n>0) {
      int received = 0;
      do {
        if (is.available() > 0) {
          i = is.read(array, i, n-received);  

          // break if connection broken
          if (i == -1 || i == 0)
            break;

          received += i;
        }
      } while (received < n);
    }

    // receive remaining parameters
    do {
      i = is.read();
    } while (i != -1 && i != 0);

    return array;
  }
  
  public String replyParam(String reply, int index) {
    String param = "";
	int i, begin, end;
	
    // find beginning of param
	for (i=begin=0 ; i<index ; i++) {
	  while (begin < reply.length() && reply.charAt(begin) != '&')
	    begin++;
	  if (reply.charAt(begin) == '&')
	    begin++;
	}
	  
	// find end of param
	end = begin;
	while (end < reply.length() && reply.charAt(end) != '&')
	  end++;
	  
	return reply.substring(begin, end);
	}
	
  private char binToAscii(int i) {
    if (i>=38)
      i = i-38+(int) 'a';
    else if (i>=12)
      i = i-12+(int) 'A';
    else
      i = i+'.';
    return (char) i;
  }
  
  private String crypt(String buf, String salt) {
    int i, seed;
    StringBuffer enc_pw = new StringBuffer(salt);

    for (i=0; i<8 && i<buf.length() ; i++)
		  enc_pw.append(buf.charAt(i));
	  for (; i<11; i++)
		  enc_pw.append((char) 0);

    seed = 123;
	  for (i=2; i<13; i++)
      {
		  seed = 5*seed + 27 + (int) enc_pw.charAt(i);
      enc_pw.setCharAt(i, binToAscii(seed & 0x3F));
      }

    return enc_pw.toString();
  }
 
  public void messageSend(int type, String msg) throws IOException {
    rpcCall("cm_msg&"+type+"&"+client+"&?&?&"+msg);
  }
  
  public void messageSend(int type, String source, String msg) throws IOException {
    rpcCall("cm_msg&"+type+"&"+client+"&?&"+source+"&"+msg);
  }

  public Vector retrieveMessages() throws IOException {
    Vector msg = new Vector();
    
    String r = rpcCall("cm_msg_retrieve&1000");
    
    // strip status
    r = r.substring(r.indexOf('&')+1, r.length());

    // copy messages into vector
    int ib = 0;
    int ie;
    
    do {
      ie = r.indexOf('\n', ib);
      if (ie > -1) {
        if (r.charAt(ie-1) == '\r') { // CR-LF msg
          msg.addElement(new String(r.substring(ib, ie-1)));
        }
        else { // LF msg
          msg.addElement(new String(r.substring(ib, ie)));
        }
        ib = ie+1;
      }
    } while (ie > -1);
    
    return msg;
  }
  
  public void addMessageListener(MessageListener ml) throws IOException {
    // open message buffer
    String r = rpcCall("bm_open_buffer&SYSMSG&100000");
    hBufMsg = Integer.parseInt(replyParam(r, 1));
    rpcCall("bm_add_event_request&"+hBufMsg+"&-1&-1&2&1&0");
    rpcCall("bm_mark_read_waiting&1");
    
    messageListener = ml;
  }
  
  public boolean isClient(String client, boolean unique) throws IOException {
    String r = rpcCall("cm_exist&"+client+"&"+(unique?"1":"0"));
    
    return replyParam(r, 0).equals("1");
  }
  
  public void disconnect() {
    try {
      if (hDB > 0) {
      
        // Send stop notification  
        messageSend(MINFO, "Program "+client+" on host "+
                    InetAddress.getLocalHost().getHostName()+" stopped");
        
        // send shutdown message
        rpcCall("id_exit");
 
        // mark experiment disconnected
        hDB = 0;
        
        // close sockets
        sockSend.close();
        sockRecv.close();
        sockEvent.close();
      }
    } catch (Exception e) {}
  }
  
  public void actionPerformed(ActionEvent e) {
    // check receive socket
    try {
      String str = "";
      int    i;
      OdbKey key;
      
      i = recvIs.read();
      
      // if we come here, no exception has been raised, therefore
      // we can increase the timeout to receive the full message
      sockRecv.setSoTimeout(1000);

      // receive message until '0' is received
      do
        {
        str += (char) i;
        i = recvIs.read();
        } while (i > 0);

      // reset timeout
      sockRecv.setSoTimeout(1);

      StringTokenizer st = new StringTokenizer(str, "&");
      String msg = st.nextToken();
      if (msg.equals("MSG_ODB")) {
        st.nextToken();
        int hKey = Integer.parseInt(st.nextToken());
        
        // lookup key
        for (i=0 ; i<vecHotLink.size() ; i++) {
          key = (OdbKey)vecHotLink.elementAt(i); 
          if (key.getHkey() == hKey) {
            key.dispatchUpdate();
            }
        }
      } else if (msg.equals("MSG_BM")) {
        String bufName = st.nextToken();

        // receive message(s)
        if (bufName.equals("SYSMSG")) {
          do {
            byte array[] = rpcReceiveEvent(hBufMsg);
            
            // break if no more events
            if (array.length == 0)
              break;
              
            // convert array (without header to string)
            String message = new String(array, 16, array.length-16);
              
            // remove padding
            int end = message.indexOf(0);
            message = message.substring(0, end);

            if (messageListener != null)
                messageListener.receiveMessage(message);
          } while (true);
        }
        
        // receive event
        else {
        }
      }
    } catch (SocketException se) { 
        JOptionPane.showMessageDialog(null, 
          "Conecction to server lost, stoppping program",
          "JMIDAS" , 
          JOptionPane.ERROR_MESSAGE, 
          null);
        System.exit(1);
    } catch (Exception ex) { }
  }

  public void addHotLink(OdbKey key) {
    vecHotLink.addElement(key);
  }

  public void removeHotLink(OdbKey key) {
    vecHotLink.removeElement(key);
  }
  
  public void start() {
    start(0);
  }
  
  public String getOdbValue(String keyName) throws OdbKeyNotFoundException, IOException {
    OdbKey key = new OdbKey(this, keyName);
    return key.toString();
  }

  public void setOdbValue(String keyName, String value) throws OdbKeyNotFoundException, IOException {
    OdbKey key = new OdbKey(this, keyName);
    key.setText(value);
  }

  public OdbKey enumOdbKey(String startPath, int index) throws OdbKeyNotFoundException, IOException {
    OdbKey startKey = new OdbKey(this, startPath);
    String r = rpcCall("db_enum_key&"+hDB+"&"+startKey.getHkey()+"&"+index);
    if (replyParam(r, 0).equals("1")) {
      int hSubkey = Integer.parseInt(replyParam(r, 1));
      OdbKey subkey = new OdbKey(this, hSubkey);
      return subkey;
    } else return null;
  }
  
  public void waitForState(int state) throws TooManyListenersException, 
                OdbKeyNotFoundException, IOException, InterruptedException
  {
    OdbKey k = new OdbKey(this, "/Runinfo/State");
    k.hotLink();
    while (Integer.parseInt(k.toString()) != state)
      Thread.sleep(1000);
  }
  
  public void start(int runNumber) {
    try {
      // retrieve current run number
      OdbKey oldRunNumber = new OdbKey(this, "/Runinfo/Run number");

      // increment run number
      if (runNumber == 0)
        runNumber = Integer.parseInt(oldRunNumber.toString())+1;
        
      String r = rpcCall("cm_transition&1&"+runNumber+"&256&0");
      
      if (!replyParam(r, 0).equals("1")) {
        // reset run number
        oldRunNumber.setText(oldRunNumber.toString());
        
        JOptionPane.showMessageDialog(null, 
          "Error starting run: "+replyParam(r, 1),
          "Error" , 
          JOptionPane.ERROR_MESSAGE, 
          null);
      }
      
    } catch (IOException e) { System.out.println(e); }
  }
  
  public void startDialog(Component parent) {
        
	try {
      // check if transition in progress
      OdbKey trans = new OdbKey(this, "/Runinfo/Transition in progress");
      if (trans.toString().equals("1"))
        JOptionPane.showMessageDialog(parent, 
          "Start/Stop already in progress, please try again later or set\n"+
          "\"/Runinfo/Transition in progress\" manually to zero.",
          "JMIDAS experiment control",
          JOptionPane.WARNING_MESSAGE, null);
      else {
        trans.setText("1");

        JPanel p[] = new JPanel[10];
        JLabel l[] = new JLabel[10];
        MJTextField  t[] = new MJTextField[10];
        
        try {
          // run number field with incremented run number
          p[0] = new JPanel(new GridLayout(1, 2));    
          l[0] = new JLabel("Run number");
          t[0] = new MJTextField("/Runinfo/Run number", false);
          t[0].connect(this);
          int rn = Integer.parseInt(t[0].getData())+1;
          t[0].setData(""+rn);
          p[0].add(l[0]);
          p[0].add(t[0]);

          // loop through experiment parameters
          OdbKey expParam;
          for (int i=0 ; ; i++) {
            expParam = enumOdbKey("/Experiment/Edit on start", i);
            if (expParam == null)
            break;
              
            p[i+1] = new JPanel(new GridLayout(1, 2));
            l[i+1] = new JLabel(expParam.getName()+":");
            t[i+1] = new MJTextField(expParam.getPath(), false);
            t[i+1].connect(this);
            p[i+1].add(l[i+1]);
            p[i+1].add(t[i+1]);
          }

        } catch (OdbKeyNotFoundException e) {};
        
        int reply = JOptionPane.showConfirmDialog(parent, 
                      p, "Enter run parameters" , 
                      JOptionPane.OK_CANCEL_OPTION, 
                      JOptionPane.PLAIN_MESSAGE);
        if (reply == JOptionPane.OK_OPTION) {
          try {
            for (int i=1 ; t[i] != null ; i++)
            t[i].propagate();
          } catch (IOException e) { System.out.println(e); }

          // start run with new run number
          start(Integer.parseInt(t[0].getData()));
        }
        
        trans.setText("0");
	  }
	} catch (Exception e) { System.out.println(e); };
  }
  
  public void stopDialog(Component parent) {
	try {
      // check if transition in progress
      OdbKey trans = new OdbKey(this, "/Runinfo/Transition in progress");
      if (trans.toString().equals("1"))
        JOptionPane.showMessageDialog(parent, 
          "Start/Stop already in progress, please try again later or set\n"+
          "\"/Runinfo/Transition in progress\" manually to zero.",
          "JMIDAS experiment control",
          JOptionPane.WARNING_MESSAGE, null);
      else {
        trans.setText("1");
	    stop();
        trans.setText("0");
	  }
	} catch (Exception e) { System.out.println(e); };
  }
	  
  public void stop() {
    try {
      String r = rpcCall("cm_transition&2&0&256&0");
      
      if (!replyParam(r, 0).equals("1")) {
        JOptionPane.showMessageDialog(null, 
          "Error stopping run: "+replyParam(r, 1),
          "Error" , 
          JOptionPane.ERROR_MESSAGE, 
          null);
      }
      
    } catch (IOException e) { System.out.println(e); }
  }

  public void pause() {
    try {
      String r = rpcCall("cm_transition&4&0&256&0");
      
      if (!replyParam(r, 0).equals("1")) {
        JOptionPane.showMessageDialog(null, 
          "Error pausing run: "+replyParam(r, 1),
          "Error" , 
          JOptionPane.ERROR_MESSAGE, 
          null);
      }
      
    } catch (IOException e) { System.out.println(e); }
  }

  public void resume() {
    try {
      String r = rpcCall("cm_transition&8&0&256&0");
      
      if (!replyParam(r, 0).equals("1")) {
        JOptionPane.showMessageDialog(null, 
          "Error resuming run: "+replyParam(r, 1),
          "Error" , 
          JOptionPane.ERROR_MESSAGE, 
          null);
      }
      
    } catch (IOException e) { System.out.println(e); }
  }

  protected void finalize() throws Exception {
    disconnect();
  }
  
}

